(*pp deriving *)

module type TypesSig =
sig
  type datatype
      deriving (Eq, Show, Pickle, Typeable, Shelve) (* , Show, Pickle, Typeable, Shelve) *)
end

module type Q =
sig
  
  type datatype
      deriving (Eq, Show, Pickle, Typeable, Shelve)(* , Show, Pickle, Typeable, Shelve) *)

  type like_expr =
      [ `caret
      | `dollar
      | `underscore
      | `percent
      | `seq of like_expr list
      | `string of string
      | `variable of string ]
  type table_spec = [ `TableName of string | `TableVariable of string ]
  type table_instance = table_spec * string
  type expression =
      Field of (string * string)
      | Variable of string
      | Null
      | Integer of Num.num
      | Float of float
      | Boolean of bool
      | LikeExpr of like_expr
      | Text of string
      | Funcall of (string * expression list)
      | Binary_op of (string * expression * expression)
      | Unary_op of (string * expression)
      | Case of (expression * expression) list
      | Query of query
  and query = {
    distinct_only : bool;
    result_cols : col_or_expr list;
    tables : table_instance list;
    condition : expression;
    sortings : sorting list;
    max_rows : expression option;
    offset : expression;
    projections : string list;
  }
  and sorting = [ `Asc of string * string | `Desc of string * string ]
  and column = {
    table_alias : string;
    name : string;
    col_alias : string;
    col_type : datatype;
  } 
  and exprcol = {
    expression : expression;
    exprcol_alias : string;
    exprcol_table_alias : string;
  }
  and col_or_expr = (column, exprcol) Utility.either
      deriving (Eq, Show, Pickle, Typeable, Shelve)(*  Typeable, Show, Pickle, Shelve) *)
      
  val like_as_string : like_expr -> string
  val owning_table : string -> query -> string
  val query_replace_var : string -> expression -> query -> query
  val add_sorting : query -> sorting -> query
  val freevars : query -> string list
  val occurs_free : string -> query -> bool

end    
  
module Make (T : TypesSig) : Q with type datatype = T.datatype

(*
  mini-sql that we handle here:
  
  SELECT [ DISTINCT * | expression [ AS output_name ] ]
  [ FROM from_item [, ...] ]
  [ WHERE condition ]
    [ ORDER BY expression [ ASC | DESC ] ]
  [ LIMIT { count } ]
  [ OFFSET start ]
  
  where from_item is:
  
  table_name [ [ AS ] alias ]
*)
