/*
 * testlibpq.c
 *
 *      Test the C version of libpq, the PostgreSQL frontend library.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>          /* for strlen */

#include <pcre.h>
#include "libpq-fe.h"

static void
exit_nicely(PGconn *conn)
{
    PQfinish(conn);
    exit(1);
}

int
main(int argc, char **argv)
{
    const char *conninfo;
    const char *keyword;
    pcre       *re;
    const char *errptr;
    int        erroffset;
    const int  ovecsize = 3;
    int        ovector[ovecsize];
    PGconn     *conn;
    PGresult   *res;
    int         nFields;
    const char *text;
    int         i,
                j;
    int         nresults;
    int         mode;           /* 1 for server, 2 for DB */
    char        query[128];
    
    /* arguments: 
       1:host
       2:port
       3:user
       4:database
       5:table
       6:keyword
       7:mode
     */

    if (argc < 8) {
        fprintf(stderr, "Usage: %s host port user database keyword mode\n", 
            argv[0]);
        exit(1);
    }
    
    keyword = argv[6];
    mode = atoi(argv[7]);
    
    re = pcre_compile(keyword, 0, &errptr, &erroffset, NULL);
    if (re == NULL) {
        fprintf(stderr, "PCRE Error: %s\n", *errptr);
        exit(1);
    }

    conninfo="user=selinks host=localhost port=5430 dbname=sewiki_conversion";

    /* Make a connection to the database */
    conn = PQsetdbLogin(
        argv[1],                /* pghost */
        argv[2],                /* pgport */
        NULL,                   /* pgoptions */
        NULL,                   /* pgtty */
        argv[4],                /* dbName */
        argv[3],                /* login */
        NULL                    /* pwd */
        );
    

    /* Check to see that the backend connection was successfully made */
    if (PQstatus(conn) != CONNECTION_OK)
    {
        fprintf(stderr, "Connection to database failed: %s",
                PQerrorMessage(conn));
        exit_nicely(conn);
    }

    switch (mode) {
        
    case 2:                /* mode 2 = db-side, no policy */

        snprintf(query, sizeof(query), 
                 "SELECT pageid, text FROM %s WHERE text LIKE '%%%s%%'", 
                 argv[5],
                 keyword);
        res = PQexec(conn, query);
        if (PQresultStatus(res) != PGRES_TUPLES_OK)
        {
            fprintf(stderr, "SELECT command failed: %s", PQerrorMessage(conn));
            PQclear(res);
            exit_nicely(conn);
        }

        nresults = 0;

        for (i = 0; i < PQntuples(res); i++)
        {
            text = PQgetvalue(res, i, 1);
            ++nresults;
        }

        break;
        

    case 1:                     /* mode 1 = server-side */
    default:

    res = PQexec(conn, "SELECT pageid, text FROM stuff");
    if (PQresultStatus(res) != PGRES_TUPLES_OK)
    {
        fprintf(stderr, "SELECT command failed: %s", PQerrorMessage(conn));
        PQclear(res);
        exit_nicely(conn);
    }

/*     /\* first, print out the attribute names *\/ */
/*     nFields = PQnfields(res); */
/*     for (i = 0; i < nFields; i++) */
/*         printf("%-15s", PQfname(res, i)); */
/*     printf("\n\n"); */

    nresults = 0;

    /* next, print out the rows */
    for (i = 0; i < PQntuples(res); i++)
    {
        text = PQgetvalue(res, i, 1);
        if (pcre_exec(re, NULL, text, strlen(text), 0, 0, ovector, ovecsize) > 0) {
/*             for (j = 0; j < nFields; j++) */
/*                 printf("%-5s", PQgetvalue(res, i, j)); */
/*             printf("\n"); */
            ++nresults;
        }
    }
    }
    
    PQclear(res);

    /* close the connection to the database and cleanup */
    PQfinish(conn);

    printf("%d\n", nresults);

    return 0;
}
